/* $Id: clip.c,v 3.15 1999/04/28 22:55:40 keithw Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */





#ifdef PC_HEADER
#include "all.h"
#else
#include <string.h>
#include <stdlib.h>
#include "clip.h"
#include "context.h"
#include "macros.h"
#include "matrix.h"
#include "types.h"
#include "vb.h"
#include "xform.h"
#ifdef XFree86Server
#include "GL/xf86glx.h"
#endif
#endif




/* Linear interpolation between A and B: */
#define LINTERP( T, A, B )   ( (A) + (T) * ( (B) - (A) ) )



#define INTERP_SZ( t, vec, to, a, b, sz )			\
do {								\
   switch (sz) {						\
   case 4: vec[to][3] = LINTERP( t, vec[a][3], vec[b][3] );	\
   case 3: vec[to][2] = LINTERP( t, vec[a][2], vec[b][2] );	\
   case 2: vec[to][1] = LINTERP( t, vec[a][1], vec[b][1] );	\
   case 1: vec[to][0] = LINTERP( t, vec[a][0], vec[b][0] );	\
   }								\
} while(0)
   

/*
 * This function is used to interpolate colors, indexes, and texture
 * coordinates when clipping has to be done.  In general, we compute
 *     aux[dst] = aux[in] + t * (aux[out] - aux[in])
 * where aux is the quantity to be interpolated.
 * Input:  dst - index of array position to store interpolated value
 *         t - a value in [0,1]
 *         in - index of array position corresponding to 'inside' vertex
 *         out - index of array position corresponding to 'outside' vertex
 */
void gl_clip_interp_all( struct vertex_buffer *VB,
                         GLuint dst, GLfloat t, GLuint in, GLuint out )
{
   GLuint clipmask = VB->ctx->ClipMask;
   if (VB->EdgeFlagPtr)
      VB->EdgeFlagPtr->data[dst] = VB->EdgeFlagPtr->data[out];
   VB->ClipMask[dst] = 0;

   if (clipmask & CLIP_FCOLOR_BIT) {
      INTERP_SZ( t, VB->Color[0]->data, dst, in, out, 4 );
   }
   else if (clipmask & CLIP_FINDEX_BIT) {
      VB->IndexPtr->data[dst] = (GLuint) (GLint) 
	 LINTERP( t, 
		  (GLfloat) VB->Index[0]->data[in],
		  (GLfloat) VB->Index[0]->data[out] );
   }

   if (clipmask & CLIP_BCOLOR_BIT) {
      INTERP_SZ( t, VB->Color[1]->data, dst, in, out, 4 );
   }
   else if (clipmask & CLIP_BINDEX_BIT) {
      VB->Index[1]->data[dst] = (GLuint) (GLint) 
	 LINTERP( t, 
		  (GLfloat) VB->Index[1]->data[in],
		  (GLfloat) VB->Index[1]->data[out] );
   }

   if (clipmask & CLIP_TEXTURE_BIT) {
      GLuint i;
      GLuint nr = VB->ctx->Const.MaxTextureUnits;
      for (i=0;i<nr;i++) {
	 INTERP_SZ( t, 
		    VB->TexCoordPtr[i]->data, 
		    dst, in, out, 
		    VB->TexCoordPtr[i]->size );
      }
   }
}


/*
 * Specialized version gl_clip_interp_all:  just do color and 1 texcoord.
 */
void gl_clip_interp_color_tex( struct vertex_buffer *VB,
                               GLuint dst, GLfloat t, GLuint in, GLuint out )
{
   GLuint tc = VB->ctx->Current.TexSet;
   if (VB->EdgeFlagPtr)
      VB->EdgeFlagPtr->data[dst] = VB->EdgeFlagPtr->data[out];
   VB->ClipMask[dst] = 0;
   INTERP_SZ( t, VB->ColorPtr->data, dst, in, out, 4 );
   INTERP_SZ( t, VB->TexCoordPtr[tc]->data, dst, in, out, 
	      VB->TexCoordPtr[tc]->size );
}


/*
 * Specialized version gl_clip_interp_all:  just do 1 texcoord.
 */
void gl_clip_interp_tex( struct vertex_buffer *VB, 
                         GLuint dst, GLfloat t, GLuint in, GLuint out )
{
   GLuint tc = VB->ctx->Current.TexSet;
   if (VB->EdgeFlagPtr)
      VB->EdgeFlagPtr->data[dst] = VB->EdgeFlagPtr->data[out];
   VB->ClipMask[dst] = 0;
   INTERP_SZ( t, VB->TexCoordPtr[tc]->data, dst, in, out, 
	      VB->TexCoordPtr[tc]->size );
}


/*
 * Specialized version gl_clip_interp_all:  just do color.
 */
void gl_clip_interp_color( struct vertex_buffer *VB, 
                           GLuint dst, GLfloat t, GLuint in, GLuint out )
{
   if (VB->EdgeFlagPtr)
      VB->EdgeFlagPtr->data[dst] = VB->EdgeFlagPtr->data[out];
   VB->ClipMask[dst] = 0;
   INTERP_SZ( t, VB->ColorPtr->data, dst, in, out, 4 );
}

void gl_clip_interp_none( struct vertex_buffer *VB, 
			  GLuint dst, GLfloat t, GLuint in, GLuint out )
{
   if (VB->EdgeFlagPtr)
      VB->EdgeFlagPtr->data[dst] = VB->EdgeFlagPtr->data[out];
   (void) t;
   (void) in;
   VB->ClipMask[dst] = 0;
}



/**********************************************************************/
/*                     Get/Set User clip-planes.                      */
/**********************************************************************/



void gl_ClipPlane( GLcontext* ctx, GLenum plane, const GLfloat *equation )
{
   GLint p;

   p = (GLint) plane - (GLint) GL_CLIP_PLANE0;
   if (p<0 || p>=MAX_CLIP_PLANES) {
      gl_error( ctx, GL_INVALID_ENUM, "glClipPlane" );
      return;
   }

   /*
    * The equation is transformed by the transpose of the inverse of the
    * current modelview matrix and stored in the resulting eye coordinates.
    *
    * KW: Eqn is then transformed to the current clip space, where user
    * clipping now takes place.  The clip-space equations are recalculated
    * whenever the projection matrix changes.
    */
   if (ctx->ModelView.flags & MAT_DIRTY_ALL_OVER) {
      gl_matrix_analyze( &ctx->ModelView );
   }
   gl_transform_vector( ctx->Transform.EyeUserPlane[p], equation,
		        ctx->ModelView.inv );

   if (ctx->Transform.ClipEnabled[p]) {
      if (ctx->ProjectionMatrix.flags & MAT_DIRTY_ALL_OVER) {
	 gl_matrix_analyze( &ctx->ProjectionMatrix );
      }
      gl_transform_vector( ctx->Transform.ClipUserPlane[p], 
			   ctx->Transform.EyeUserPlane[p], 
			   ctx->ProjectionMatrix.inv );
   }
}


void gl_update_userclip( GLcontext *ctx )
{
   GLuint p;
   
   for (p = 0 ; p < MAX_CLIP_PLANES ; p++) {
      if (ctx->Transform.ClipEnabled[p]) {
	 gl_transform_vector( ctx->Transform.ClipUserPlane[p],
			      ctx->Transform.EyeUserPlane[p],
			      ctx->ProjectionMatrix.inv );
      }
   }
}

void gl_GetClipPlane( GLcontext* ctx, GLenum plane, GLdouble *equation )
{
   GLint p;

   ASSERT_OUTSIDE_BEGIN_END(ctx, "glGetClipPlane");


   p = (GLint) (plane - GL_CLIP_PLANE0);
   if (p<0 || p>=MAX_CLIP_PLANES) {
      gl_error( ctx, GL_INVALID_ENUM, "glGetClipPlane" );
      return;
   }

   equation[0] = (GLdouble) ctx->Transform.EyeUserPlane[p][0];
   equation[1] = (GLdouble) ctx->Transform.EyeUserPlane[p][1];
   equation[2] = (GLdouble) ctx->Transform.EyeUserPlane[p][2];
   equation[3] = (GLdouble) ctx->Transform.EyeUserPlane[p][3];
}




/**********************************************************************/
/*                         View volume clipping.                      */
/**********************************************************************/


/*
 * Clip a point against the view volume.
 * Input:  v - vertex-vector describing the point to clip
 * Return:  0 = outside view volume
 *          1 = inside view volume
 */
GLuint gl_viewclip_point( const GLfloat v[] )
{
   if (   v[0] > v[3] || v[0] < -v[3]
       || v[1] > v[3] || v[1] < -v[3]
       || v[2] > v[3] || v[2] < -v[3] ) {
      return 0;
   }
   else {
      return 1;
   }
}

/*
 * Clip a point against the user clipping planes.
 * Input:  v - vertex-vector describing the point to clip.
 * Return:  0 = point was clipped
 *          1 = point not clipped
 */
GLuint gl_userclip_point( GLcontext* ctx, const GLfloat v[] )
{
   GLuint p;

   for (p=0;p<MAX_CLIP_PLANES;p++) {
      if (ctx->Transform.ClipEnabled[p]) {
	 GLfloat dot = v[0] * ctx->Transform.ClipUserPlane[p][0]
		     + v[1] * ctx->Transform.ClipUserPlane[p][1]
		     + v[2] * ctx->Transform.ClipUserPlane[p][2]
		     + v[3] * ctx->Transform.ClipUserPlane[p][3];
         if (dot < 0.0F) {
            return 0;
         }
      }
   }

   return 1;
}





clip_poly_func gl_poly_clip_tab[5];
clip_line_func gl_line_clip_tab[5];


#if defined(__i386__)
#define NEGATIVE(x) ((*(int *)&x)<0)
#else
#define NEGATIVE(x) (x < 0)
#endif

#define W(i) coord[i][3]
#define Z(i) coord[i][2]
#define Y(i) coord[i][1]
#define X(i) coord[i][0]
#define SIZE 4
#define TAG(x) x##_4
#include "clip_funcs.h"


#define W(i) 1.0
#define Z(i) coord[i][2]
#define Y(i) coord[i][1]
#define X(i) coord[i][0]
#define SIZE 3
#define TAG(x) x##_3
#include "clip_funcs.h"

#define W(i) 1.0
#define Z(i) 0.0
#define Y(i) coord[i][1]
#define X(i) coord[i][0]
#define SIZE 2
#define TAG(x) x##_2
#include "clip_funcs.h"



GLboolean gl_user_cliptest( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   GLvector4f *coord = VB->ClipPtr;
   GLvector4f out;
   GLubyte *clipMask = VB->ClipMask + VB->Start; 
   GLuint p, i;

   if (!VB->tmp_m) 
      VB->tmp_m = (GLfloat *)malloc(VB->Size * sizeof(GLfloat));

   out.data = (GLfloat (*)[4])VB->tmp_m;
   out.start = VB->tmp_m;
   out.count = coord->count;
   out.stride = sizeof(GLfloat);

   for (p = 0; p < MAX_CLIP_PLANES ; p++)
   {
      GLuint nr = 0;

      if (!ctx->Transform.ClipEnabled[p])
	 continue;
      
      gl_dotprod_tab[0][coord->size]( &out, 0, coord, 
				      ctx->Transform.ClipUserPlane[p], 0 );

      for (i = 0 ; i < out.count ; i++) {
	 if (NEGATIVE(out.start[i])) {
	    clipMask[i] |= CLIP_USER_BIT;
	    nr++;
	 }
      }

      if (nr == 0) continue;
      VB->ClipOrMask |= CLIP_USER_BIT;
      VB->CullMode |= CLIP_MASK_ACTIVE;
      if (nr == out.count) return GL_FALSE;
   }

   return GL_TRUE;
}
		       

void gl_init_clip(void)
{
   init_clip_funcs_4();
   init_clip_funcs_3();
   init_clip_funcs_2();
}

