/* $Id: cva.c,v 3.8 1999/04/25 10:18:38 keithw Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* Mesa CVA implementation.  
 * Copyright (C) 1999 Keith Whitwell
 */


#include "api.h"
#include "types.h"
#include "cva.h"
#include "context.h"
#include "pipeline.h"
#include "varray.h"
#include "vbcull.h"
#include "vbrender.h"
#include "vbxform.h"
#include "vector.h"
#include <stdlib.h>


static void copy_clipmask( GLubyte *dest, GLubyte *ormask, GLubyte *andmask,
			   const GLubyte *src, 
			   const GLuint *elt, GLuint nr )
{
   GLuint i;
   GLubyte o = *ormask;
   GLubyte a = *andmask;

   for (i = 0 ; i < nr ; i++) {
      GLubyte t = src[elt[i]];
      dest[i] = t;
      o |= t;
      a &= t;
   }  

   *ormask = o;
   *andmask = a;
}

static void translate_4f( GLvector4f *dest, 
			  CONST GLvector4f *src, 
			  CONST GLuint elt[], 
			  GLuint nr )
{
   GLuint i;
   GLfloat (*from)[4] = (GLfloat (*)[4])src->start;
   GLfloat (*to)[4] = (GLfloat (*)[4])dest->start;
   GLuint stride = src->stride;
   
   if (stride == 4 * sizeof(GLfloat)) {
      for (i = 0 ; i < nr ; i++) 
	 COPY_4FV( to[i], from[elt[i]] );
   } else {
      for (i = 0 ; i < nr ; i++) {
	 CONST GLubyte *f = (GLubyte *)from + elt[i] * stride;
	 COPY_4FV( to[i], (GLfloat *)f );
      }
   }      

   dest->size = src->size;
   dest->flags |= (src->flags & VEC_SIZE_4);
   dest->count = nr;
}

static void translate_3f( GLvector3f *dest, 
			  CONST GLvector3f *src, 
			  CONST GLuint elt[], 
			  GLuint nr )
{
   GLuint i;
   GLfloat (*from)[3] = (GLfloat (*)[3])src->start;
   GLfloat (*to)[3] = (GLfloat (*)[3])dest->start;
   GLuint stride = src->stride;

   if (stride == 3 * sizeof(GLfloat)) {
      for (i = 0 ; i < nr ; i++) 
	 COPY_3FV( to[i], from[elt[i]] );
   } else {
      for (i = 0 ; i < nr ; i++) {
	 CONST GLubyte *f = (GLubyte *)from + elt[i] * stride;
	 COPY_3FV( to[i], (GLfloat *)f );
      }
   }

   dest->count = nr;
}

static void translate_4ub( GLvector4ub *dest, 
			   CONST GLvector4ub *src,
			   GLuint elt[], 
			   GLuint nr )
{
   GLuint i;
   GLubyte (*from)[4] = (GLubyte (*)[4])src->start;
   GLubyte (*to)[4] = (GLubyte (*)[4])dest->start;
   GLuint stride = src->stride;

   if (stride == 4 * sizeof(GLubyte)) {
      for (i = 0 ; i < nr ; i++)
	 COPY_4UBV( to[i], from[elt[i]]);
   } else {
      for (i = 0 ; i < nr ; i++) {
	 CONST GLubyte *f = (GLubyte *)from + elt[i] * stride;
	 COPY_4UBV( to[i], f );
      }
   }

   dest->count = nr;
}

static void translate_1ui( GLvector1ui *dest, 
			   GLvector1ui *src, 
			   GLuint elt[], 
			   GLuint nr )
{
   GLuint i;
   GLuint *from = src->start;
   GLuint *to = dest->start;
   GLuint stride = src->stride;

   if (stride == sizeof(GLuint)) {
      for (i = 0 ; i < nr ; i++)
	 to[i] = from[elt[i]];
   } else {
      for (i = 0 ; i < nr ; i++) {
	 CONST GLubyte *f = (GLubyte *)from + elt[i] * stride;
	 to[i] = *(GLuint *)f;
      }
   }

   dest->count = nr;
}

static void translate_1ub( GLvector1ub *dest, 
			   GLvector1ub *src, 
			   GLuint elt[], 
			   GLuint nr )
{
   GLuint i;
   GLubyte *from = src->start;
   GLubyte *to = dest->start;
   GLuint stride = src->stride;

   if (stride == sizeof(GLubyte)) {
      for (i = 0 ; i < nr ; i++)
	 to[i] = from[elt[i]];
   } else {
      for (i = 0 ; i < nr ; i++) {
	 CONST GLubyte *f = from + elt[i] * stride;
	 to[i] = *f;
      }
   }

   dest->count = nr;
}


/* The fallback case for handling the merge of cva and immediate
 * data.  This code back-copies data from cva->immediate to build a
 * normal looking immediate struct, which is then sent off to the
 * old raster setup and render routines.
 *
 * The FX driver implements a replacement for this step which merges
 * the new data into the already prepared GrVertex structs.
 *
 * When there is no data to merge, gl_render_vb_indirect is called
 * instead.
 */
void gl_merge_cva( struct vertex_buffer *VB,
		   struct vertex_buffer *cvaVB )
{
   GLcontext *ctx = VB->ctx;
   GLuint *elt = VB->Elt + VB_START;
   GLuint count = VB->Count - VB_START;
   GLuint flags = ctx->CVA.pre.outputs & ctx->CVA.elt.inputs;
   
/*     gl_print_vert_flags("cva merge", flags); */

   if (flags & VERT_WIN) {
      VB->ClipPtr = &VB->Clip;
      VB->Projected = &VB->Win;
      VB->CullMode = 0;
   
      if (cvaVB->ClipOrMask) {

	 /* Copy clipmask back into VB, build a new clipOrMask */
	 copy_clipmask( VB->ClipMask + VB_START, 
			&VB->ClipOrMask, &VB->ClipAndMask,
			cvaVB->ClipMask, 
			VB->Elt + VB_START, 
			VB->Count - VB_START );
	 
	 /* overkill if !VB->ClipOrMask - should just copy 'copied' verts */
	 translate_4f( VB->ClipPtr, cvaVB->ClipPtr, elt, count);

	 if (VB->ClipOrMask) 
	    VB->CullMode |= CLIP_MASK_ACTIVE;      

	 if (VB->ClipAndMask) {
	    VB->Culled = 1;
	    gl_dont_cull_vb( VB );   
	    return;
	 }
      }

      translate_4f( &VB->Win, &cvaVB->Win, elt, count );

      /* Can't be precomputed - but may be wasteful to do this now.
       */
      if (VB->EarlyCull && (VB->ClipOrMask || ctx->NeedCull))
      {
	 GLuint cullcount = gl_cull_vb( VB );      
	 if (cullcount) VB->CullMode |= CULL_MASK_ACTIVE;
	 if (cullcount == VB->Count) { VB->Culled = 2 ; return; }
      }
      else 
	 gl_dont_cull_vb( VB );   
   } else {
      VB->ClipPtr = &VB->Clip;
      VB->Projected = &VB->Win;
      gl_dont_cull_vb( VB );   
      if (!flags) return;
   }

   if (flags & VERT_EYE) 
   {
      VB->Unprojected = VB->EyePtr = &VB->Eye;
      translate_4f( VB->EyePtr, cvaVB->EyePtr, elt, count);
   }

   if (flags & VERT_OBJ_ANY) 
   {
      VB->ObjPtr = &VB->IM->v.Obj;
      if (!ctx->NeedEyeCoords) VB->Unprojected = VB->ObjPtr;
      translate_4f( VB->ObjPtr, cvaVB->ObjPtr, elt, count);
   }

   if (flags & VERT_NORM) 
   {
      VB->NormalPtr = &VB->IM->v.Normal;      
      translate_3f( VB->NormalPtr, cvaVB->NormalPtr, elt, count );
      VB->CullMode &= ~COMPACTED_NORMALS;
   }

   if (flags & VERT_RGBA) 
   {
      VB->ColorPtr = VB->Color[0] = VB->LitColor[0];
      translate_4ub( VB->Color[0], cvaVB->Color[0], elt, count );

      if (ctx->LightTwoSide) {
	 VB->Color[1] = VB->LitColor[1];
	 translate_4ub( VB->Color[1], cvaVB->Color[1], elt, count );
      }
   }
   
   if (flags & VERT_INDEX) 
   {
      VB->IndexPtr = VB->Index[0] = VB->LitIndex[0];
      translate_1ui( VB->Index[0], cvaVB->Index[0], elt, count );

      if (ctx->LightTwoSide) {
	 VB->Index[1] = VB->LitIndex[1];
	 translate_1ui( VB->Index[1], cvaVB->Index[1], elt, count );
      }
   }

   if (flags & VERT_EDGE) 
   {
      VB->EdgeFlagPtr = &VB->IM->v.EdgeFlag;      
      translate_1ub( VB->EdgeFlagPtr, cvaVB->EdgeFlagPtr, elt, count );
   }

   if (flags & VERT_TEX0_ANY) 
   {
      VB->TexCoordPtr[0] = &VB->IM->v.TexCoord[0];
      translate_4f( VB->TexCoordPtr[0], cvaVB->TexCoordPtr[0], elt, count);
   }

   if (flags & VERT_TEX1_ANY) 
   {
      VB->TexCoordPtr[1] = &VB->IM->v.TexCoord[1];      
      translate_4f( VB->TexCoordPtr[1], cvaVB->TexCoordPtr[1], elt, count);
   }
}



#define RESET_VEC(v, t, s) (v.start = t v.data[s], v.count = 0)  

/* We don't have a good mechanism for dealing with the situation where
 * cva is 'abandoned' midway through a vertex buffer, or indeed any mixing
 * of cva & standard data in a single immediate struct.
 */
void gl_rescue_cva( GLcontext *ctx, struct immediate *IM )
{
   struct vertex_buffer *VB = ctx->VB;
   GLuint i;

   /* Basically just abandon CVA even though the precalced data
    * is already there.
    */
   IM->Start = VB->CopyStart;
   RESET_VEC(IM->v.Obj, (GLfloat *), VB->CopyStart);
   RESET_VEC(IM->v.Normal, (GLfloat *), VB->CopyStart);
   RESET_VEC(IM->v.TexCoord[0], (GLfloat *), VB->CopyStart);
   RESET_VEC(IM->v.TexCoord[1], (GLfloat *), VB->CopyStart);
   RESET_VEC(IM->v.Index, &, VB->CopyStart);
   RESET_VEC(IM->v.Color, (GLubyte *), VB->CopyStart);
   RESET_VEC(VB->Clip, (GLfloat *), VB->CopyStart);
   RESET_VEC(VB->Eye, (GLfloat *), VB->CopyStart);
   RESET_VEC(VB->Win, (GLfloat *), VB->CopyStart);
   RESET_VEC(VB->BColor, (GLubyte *), VB->CopyStart); 
   RESET_VEC(VB->BIndex, &, VB->CopyStart);
   
   for (i = IM->Start ; i < VB_START ; i++)
      IM->Flag[i] = IM->ArrayOrFlags;   

   /* overkill
    */
   ctx->CompileCVAFlag = 0;
}





/* Transform the array components now, upto the setup call.  When
 * actual draw commands arrive, the data will be merged prior to
 * calling render_vb.  
 */
void gl_LockArrays( GLcontext *ctx, GLuint first, GLuint count )
{
   ASSERT_OUTSIDE_BEGIN_END_AND_FLUSH( ctx, "unlock arrays" );

   /* Can't mix locked & unlocked - if count is too large, just
    * unlock.  
    */
   if (count <= ctx->Const.MaxArrayLockSize)
   {   
      if (!ctx->Array.LockCount)
	 ctx->Array.NewArrayState = ~0;

      ctx->Array.LockFirst = first;
      ctx->Array.LockCount = count;     
      ctx->NewState |= NEW_CLIENT_STATE;
      ctx->CompileCVAFlag = !ctx->CompileFlag;
/*        ctx->CVA.orflag = 0; */
   } 
   else
   {
      if (ctx->Array.LockCount) 
	 ctx->NewState |= NEW_CLIENT_STATE;

      ctx->Array.LockFirst = 0;
      ctx->Array.LockCount = 0;
      ctx->CompileCVAFlag = 0;
   }
}


void gl_UnlockArrays( GLcontext *ctx )
{
   ASSERT_OUTSIDE_BEGIN_END_AND_FLUSH( ctx, "unlock arrays" );
   
   ctx->Array.LockFirst = 0;
   ctx->Array.LockCount = 0;
   ctx->CompileCVAFlag = 0;
}



void gl_alloc_cva_store( struct gl_cva *cva, GLuint size )
{
   cva->store.Obj = (GLfloat (*)[4])malloc( sizeof(GLfloat) * 4 * size );
   cva->store.Normal = (GLfloat (*)[3])malloc( sizeof(GLfloat) * 3 * size );
   cva->store.Color = (GLubyte (*)[4])malloc( sizeof(GLubyte) * 4 * size );
   cva->store.EdgeFlag = (GLubyte *)malloc( sizeof(GLubyte) * size );
   cva->store.Index = (GLuint *)malloc( sizeof(GLuint) * size );
   cva->store.TexCoord[0] = (GLfloat (*)[4])malloc( sizeof(GLfloat) * 4 * size);
   cva->store.TexCoord[1] = (GLfloat (*)[4])malloc( sizeof(GLfloat) * 4 * size);
}

void gl_free_cva_store( struct gl_cva *cva )
{
   free( cva->store.Obj );
   free( cva->store.Normal );
   free( cva->store.Color );
   free( cva->store.EdgeFlag );
   free( cva->store.Index );
   free( cva->store.TexCoord[0] );
   free( cva->store.TexCoord[1] );
}


/* For cva and drawElements/drawElementsRange if the number of
 * elements is significantly greater than the size of the array, but
 * less than the max lock size.
 */
void gl_prepare_arrays_cva( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   struct gl_cva *cva = &ctx->CVA;
   GLuint start = ctx->Array.LockFirst;
   GLuint n = ctx->Array.LockCount;
   GLuint flag = cva->pre.new_inputs & ctx->Array.Summary;
   GLuint i;


   VB->Flag[VB->Count] &= ~VERT_END_VB;
   VB->Culled = 0;

/*    gl_print_vert_flags("cva import", flag); */

   if (flag & VERT_RGBA) {
      GLvector4ub *col = VB->ColorPtr = &cva->v.Color;
      VB->Color[0] = VB->Color[1] = VB->ColorPtr;
      if (ctx->Array.Color.Type == GL_UNSIGNED_BYTE && 
	  ctx->Array.Color.Size == 4 && 
	  ctx->Array.Color.StrideB == 4) 
      {
	 col->data = (GLubyte (*)[4]) ctx->Array.Color.Ptr;
	 col->stride = 4 * sizeof(GLubyte);
	 col->flags = 0;
	 col->count = n;
      } else {
	 col->data = cva->store.Color;
	 ctx->Array.ColorFunc( col->data, &ctx->Array.Color, start, n );
      }	
      col->start = VEC_ELT(col, GLubyte, start); 
   }
   
   if (flag & VERT_INDEX) {
      GLvector1ui *index = VB->IndexPtr = &cva->v.Index; 
      VB->Index[0] = VB->Index[1] = VB->IndexPtr;

      if (ctx->Array.Index.Type == GL_UNSIGNED_INT &&
	  ctx->Array.Index.StrideB == sizeof(GLuint))
      {
	 index->data = (GLuint *) ctx->Array.Index.Ptr;
	 index->stride = sizeof(GLuint);
	 index->flags = 0;
	 index->count = n;
      } else {
	 index->data = cva->store.Index;
	 ctx->Array.IndexFunc( index->data, &ctx->Array.Index, start, n );
      }	
      index->start = VEC_ELT(index, GLuint, start); 
   }

   for (i = 0 ; i < ctx->Const.MaxTextureUnits ; i++)
      if (flag & VERT_TEX_ANY(i)) {
	 GLvector4f *tc = VB->TexCoordPtr[i] = &cva->v.TexCoord[i];

	 if (ctx->Array.TexCoord[i].Type == GL_FLOAT)
	 {
	    tc->data = (GLfloat (*)[4]) ctx->Array.TexCoord[i].Ptr;
	    tc->stride = ctx->Array.TexCoord[i].StrideB;
	 } else {
	    tc->data = cva->store.TexCoord[i];
	    ctx->Array.TexCoordFunc[i]( tc->data, 
					&ctx->Array.TexCoord[i], start, n );
	 }		 
	 tc->flags = 0;
	 tc->count = n;
	 tc->start = VEC_ELT(tc, GLfloat, start); 
	 tc->size = ctx->Array.TexCoord[i].Size;
      }
      
   if (flag & VERT_OBJ_ANY) {
      GLvector4f *obj = VB->ObjPtr = &cva->v.Obj;

      if (ctx->Array.Vertex.Type == GL_FLOAT)
      {
	 obj->flags = 0;
	 obj->data = (GLfloat (*)[4]) ctx->Array.Vertex.Ptr;
	 obj->stride = ctx->Array.Vertex.StrideB;
	 obj->size = ctx->Array.Vertex.Size;
	 obj->count = n;
      } else {
	 obj->data = cva->store.Obj;
	 ctx->Array.VertexFunc( obj->data, &ctx->Array.Vertex, start, n );
      }
      obj->start = VEC_ELT(obj, GLfloat, start);
   }

   if (flag & VERT_NORM) {
      GLvector3f *norm = VB->NormalPtr = &cva->v.Normal;
      if (ctx->Array.Normal.Type == GL_FLOAT) 
      {
	 norm->flags = 0;
	 norm->data = (GLfloat (*)[3]) ctx->Array.Normal.Ptr;
	 norm->stride = ctx->Array.Normal.StrideB;
	 norm->count = n;
	 norm->start = VEC_ELT(norm, GLfloat, start);
      } else {
	 norm->data = cva->store.Normal;
	 ctx->Array.NormalFunc( norm->data, &ctx->Array.Normal, start, n );
      }
    }

   if (flag & VERT_EDGE) {
      GLvector1ub *edge = VB->EdgeFlagPtr = &cva->v.EdgeFlag;

      if (ctx->Array.EdgeFlag.StrideB == sizeof(GLubyte)) {
	 edge->flags = 0;
	 edge->data = (GLboolean *) ctx->Array.EdgeFlag.Ptr;
	 edge->stride = sizeof(GLboolean);
	 edge->count = n;
	 edge->start = VEC_ELT(edge, GLboolean, start);
      } else {
	 edge->data = cva->store.EdgeFlag;
	 ctx->Array.EdgeFlagFunc( edge->data, &ctx->Array.EdgeFlag, start, n );
      }
   }

   /* This is a little annoying - these values are pretty redundant in
    * this setting, but still have to keep them uptodate.
    */
   if (ctx->Array.Flags != VB->Flag[0])
      VB->FlagMax = 0;

   if (VB->FlagMax < n) {
      for (i = VB->FlagMax ; i < n ; i++) 
	 VB->Flag[i] = ctx->Array.Flags;
      VB->Flag[i] = 0;
      VB->FlagMax = n;
   }

   VB->Flag[n] |= VERT_END_VB;
   VB->Count = n;
}

void gl_cva_force_precalc( GLcontext *ctx )
{
   struct gl_cva *cva = &ctx->CVA;

   gl_run_pipeline( cva->VB, PIPE_CVA_RUN_PRECALC );
   cva->pre.data_valid = 1;
}


void gl_flush_cva( GLcontext *ctx )
{
   struct gl_cva *cva = &ctx->CVA;
   GLuint i;

   if (cva->queue == CVA_QUEUE_SIZE || cva->count*2 > ctx->Array.LockCount*3) 
/*    if (cva->count) */
   {
      if (!cva->pre.pipeline_valid) 
	 gl_update_cva_pipelines( ctx );
   
      if (!cva->pre.data_valid)
	 gl_cva_force_precalc( ctx );
   }

   if (cva->pre.data_valid) {
      for (i = 0 ; i < cva->queue ; i++) 
	 gl_execute_cassette( ctx, cva->IMQ[i], PIPE_CVA_ELT );      
   } else {
      for (i = 0 ; i < cva->queue ; i++) {
	 gl_exec_array_elements( ctx, cva->IMQ[i] );
	 gl_execute_cassette( ctx, cva->IMQ[i], PIPE_IMMEDIATE );      
      }
   }

   if (ctx->input->ref_count == 0) 
      gl_immediate_free( ctx->input );
   
   SET_IMMEDIATE( ctx, ctx->VB->IM );
   gl_reset_input( ctx );

   cva->count = 0;
   cva->queue = 0;
/*     cva->orflag = 0; */
}


/* Queues up a few buffers before firing the cva precalc.  This should
 * help to avoid problems with precalculating lighting & then getting
 * a glMaterial() call...
 */
void gl_cva_compile_cassette( GLcontext *ctx )
{
   struct immediate *IM = ctx->input;   
   struct gl_cva *cva = &ctx->CVA;

   if ((IM->OrFlag & cva->pre.forbidden_inputs)) 
   {
      if (cva->queue) gl_flush_cva( ctx );
      cva->pre.pipeline_valid = 0;
      cva->pre.data_valid = 0;
      cva->pre.forbidden_inputs = 0;
   }

   if (cva->pre.data_valid) {
      gl_execute_cassette( ctx, IM, PIPE_CVA_ELT );
      return;
   }

   cva->orflag |= IM->OrFlag;
   cva->count += IM->Count - VB_START;
   cva->IMQ[cva->queue++] = IM;

   if (cva->queue == CVA_QUEUE_SIZE) {
      gl_flush_cva( ctx );
   } else {
      struct immediate *new_im = gl_immediate_alloc(ctx);

      if (!new_im) {
	 gl_flush_cva( ctx );
	 ctx->CompileCVAFlag = 0;
	 return;
      }

      new_im->ref_count = 0;	
      SET_IMMEDIATE( ctx, new_im );
   }
}

