/* $Id: pipeline.c,v 3.14 1999/05/16 17:09:59 keithw Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* Dynamic pipelines, support for CVA.
 * Copyright (C) 1999 Keith Whitwell.
 */


#include "bbox.h"
#include "clip.h"
#include "context.h"
#include "cva.h"
#include "pipeline.h"
#include "vbcull.h"
#include "vbindirect.h"
#include "vbrender.h"
#include "vbxform.h"
#include "fog.h"
#include "light.h"
#include "mmath.h"
#include "shade.h"
#include "types.h"
#include "xform.h"
#include <stdio.h>


GLmatrix gl_identity_mat;


void gl_init_vbxform()
{
   gl_matrix_ctr( &gl_identity_mat );
}



/* Have to reset only those parts of the vb which are being recalculated.
 */
void gl_reset_cva_vb( struct vertex_buffer *VB, GLuint stages )
{
   GLcontext *ctx = VB->ctx;

   if (stages & PIPE_OP_VERT_XFORM) {

      if (VB->CullMode & (CULL_MASK_ACTIVE|CLIP_MASK_ACTIVE)) {
	 MEMSET(VB->ClipMask, 0, VB->Count);	
      }
 
      VB->ClipOrMask = 0;
      VB->ClipAndMask = CLIP_ALL_BITS;
      VB->CullMode = 0;
      VB->Culled = 0;
      VB->LastCount = VB->Count;
   }

   if (stages & PIPE_OP_NORM_XFORM) 
      VB->NormalPtr = &ctx->CVA.v.Normal;
   
   if (stages & PIPE_OP_LIGHT) 
   {
      VB->Color[0] = VB->Color[1] = &ctx->CVA.v.Color;
      VB->Index[0] = VB->Index[1] = &ctx->CVA.v.Index;
   }
   else if (stages & PIPE_OP_FOG) 
   {
      if (ctx->Light.Enabled) {
	 VB->Color[0] = VB->LitColor[0];
	 VB->Color[1] = VB->LitColor[1];      
	 VB->Index[0] = VB->LitIndex[0];
	 VB->Index[1] = VB->LitIndex[1];      
      } else {
	 VB->Color[0] = VB->Color[1] = &ctx->CVA.v.Color;
	 VB->Index[0] = VB->Index[1] = &ctx->CVA.v.Index;
      }
   }
}


/* KW: Even if all the vertices are clipped or culled, still need to
 *     execute the material changes.
 * 
 *     TODO: Do this backwards, from count to start.  
 */
void gl_update_materials( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   GLuint orflag = VB->OrFlag;

   if (orflag & VERT_MATERIAL) 
   {
      GLuint i;
      GLuint cm_flag, interesting;
      GLubyte (*CMcolor)[4];
      GLuint *flag = VB->Flag;
      struct gl_material (*new_material)[2] = VB->Material;
      GLuint *new_material_mask = VB->MaterialMask;

      if (ctx->Light.ColorMaterialEnabled) {
	 cm_flag = VERT_RGBA;
	 interesting = VERT_END_VB | VERT_RGBA | VERT_MATERIAL;
	 CMcolor = (GLubyte (*)[4])VB->ColorPtr->start - VB->Start;
      }
      else 
      {
	 CMcolor = 0;
	 cm_flag = 0;
	 interesting = VERT_END_VB | VERT_MATERIAL;
      }

      for ( i = VB->Start ; i <= VB->Count ; i++ ) 
      {
	 while (!(flag[i] & interesting)) i++;
	 
	 if ( flag[i] & cm_flag ) 
	    gl_update_color_material( ctx, CMcolor[i] );
	 
	 if ( flag[i] & VERT_MATERIAL )
	    gl_update_material( ctx, new_material[i], new_material_mask[i] );
      }
   }
   else if ((orflag & VERT_RGBA) && ctx->Light.ColorMaterialEnabled) 
      gl_update_color_material( ctx, ctx->Current.ByteColor );   
}


/* Todo: work this out in advance.
 */
static void import_outstanding_data( struct vertex_buffer *VB )
{
   GLuint i;
   GLuint flags = VB->ctx->Flags;

   /* Can happen if ModelView and Projection matrices are both the
    * identity - GL is just the rasterizer.  Unfortunately we need to
    * copy the data for clipping or because a primitive overlapped the
    * end of the VB last time around.
    */
   if (!(VB->ClipPtr->flags & VEC_WRITABLE)) {      
      VB->ClipPtr = Transform( &VB->Clip,
			       &gl_identity_mat,
			       VB->ClipPtr,
			       VB->ClipMask + VB->Start,
			       VB->CullFlag[0]);
   }

   if ((flags & VERT_RGBA) &&
       !(VB->ColorPtr->flags & VEC_WRITABLE)) 
   {
      GLuint stride = VB->ColorPtr->stride;
      GLubyte *ptr = (GLubyte *)VB->ColorPtr->start;
      GLubyte (*dest)[4] = VB->LitColor[0]->data;
      for (i = VB->Start ; i < VB->Count ; i++, ptr += stride) {
	 COPY_4UBV( dest[i], ptr );
      }
      VB->ColorPtr = VB->LitColor[0];
   }

   if ((flags & VERT_INDEX) &&
       !(VB->IndexPtr->flags & VEC_WRITABLE)) 
   {
      GLuint stride = VB->IndexPtr->stride;
      GLuint *ptr = VB->IndexPtr->start;
      GLuint *dest = VB->store.Index->data;
      for (i = VB->Start ; i < VB->Count ; i++, 
	      ptr = (GLuint *)((GLubyte*)ptr + stride)) {
	 dest[i] = *ptr;
      }
      VB->IndexPtr = VB->store.Index;
   }

   if ((flags & VERT_TEX0_ANY) &&
       !(VB->TexCoordPtr[0]->flags & VEC_WRITABLE)) 
   {
      VB->TexCoordPtr[0] = Transform(VB->store.TexCoord[0],
				     &gl_identity_mat, 
				     VB->TexCoordPtr[0],
				     VB->ClipMask + VB->Start,
				     VB->CullFlag[0]);      
   }

   if ((flags & VERT_TEX1_ANY) && 
       !(VB->TexCoordPtr[1]->flags & VEC_WRITABLE)) 
   {
      VB->TexCoordPtr[1] = Transform(VB->store.TexCoord[1],
				     &gl_identity_mat, 
				     VB->TexCoordPtr[1],
				     VB->ClipMask + VB->Start,
				     VB->CullFlag[0]);      
   }
}




#if 0
/* Figure this is too rare to update via UpdateState, so give it a
 * flag of its own.  
 */
static void gl_calculate_model_project_win_matrix( GLcontext *ctx )
{
    gl_matrix_mul( &ctx->ModelProjectWinMatrix,
		   &ctx->Viewport.WindowMap,
		   &ctx->ModelProjectMatrix );

   gl_matrix_analyze( &ctx->ModelProjectWinMatrix );
   ctx->ModelProjectWinMatrixUptodate = GL_TRUE;
}
#endif



static void bound_cull_vb( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   VB->Projected = &VB->Win;
   VB->ClipPtr = &VB->Clip;
   VB->Culled = 1;

/*    printf("bound cull vb\n"); */

   gl_dont_cull_vb( VB );

   {
      GLuint start = 3 - VB->CopyCount;
      GLuint dst;
      GLuint *copy = VB->Copy;
      GLmatrix *mat = &ctx->ModelProjectMatrix;
      
      for (dst = start ; dst < VB->Start ; dst++) {
	 GLfloat *src = VEC_ELT(VB->ObjPtr, GLfloat, copy[dst]);
	 GLfloat *clip = VB->Clip.data[copy[dst]];
	 gl_transform_point_sz( clip, mat->m, src, VB->ObjPtr->size );
	 {
	    const GLfloat cw = clip[3];
	    const GLfloat cx = clip[0];
	    const GLfloat cy = clip[1];
	    const GLfloat cz = clip[2]; 
	    GLuint mask = 0;
	    if (cx >  cw) mask |= CLIP_RIGHT_BIT;
	    if (cx < -cw) mask |= CLIP_LEFT_BIT;
	    if (cy >  cw) mask |= CLIP_TOP_BIT;
	    if (cy < -cw) mask |= CLIP_BOTTOM_BIT;
	    if (cz >  cw) mask |= CLIP_FAR_BIT;
	    if (cz < -cw) mask |= CLIP_NEAR_BIT;
	    VB->ClipMask[copy[dst]] = mask;
	 }
      }
   }
}





static void do_vertex_pipeline( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   GLmatrix *proj_mat = ctx->vb_proj_matrix;
   GLvector4f *proj_dest = &VB->Clip;
   GLboolean need_window_transform = (1 || ctx->DoViewportMapping);
   GLubyte needclip;
   GLuint copycount = VB->CopyCount;

#define ADJ_VEC_STRIDE(v, c) (STRIDE_F((v)->start, -(c)*(v)->stride), (v)->count+=(c))


   /* Eye transform
    */
   VB->Unprojected = VB->ObjPtr;

   if (ctx->NeedEyeCoords && ctx->ModelView.type != MATRIX_IDENTITY)
      VB->Unprojected = TransformRaw( &VB->Eye, &ctx->ModelView, VB->ObjPtr );

   VB->EyePtr = VB->Unprojected;



   /* Bounding box check - never in cva or immediate.  Simulate a 
    * bounding box inclusion if clipping is disabled.
    */
   needclip = ~0;      
   if (0 && VB->ctx->Hint.ClipVolumeClipping != GL_FASTEST) {

      needclip = ~0;

      if (VB->BoundsPtr) {
	 GLubyte andmask = VB->ClipAndMask, ormask = VB->ClipOrMask;

	 gl_test_bound_tab[VB->ObjPtr->size]( &ormask, &andmask, 
					      &ctx->ModelProjectMatrix, 
					      VB->BoundsPtr );

       	 if (andmask) {
	    bound_cull_vb( VB );
	    gl_update_materials( VB );
	    return;
	 }
	 
	 needclip = ormask;

#if 0
	 /* Maybe combine the window transform into the clip transform.
	  */
	 if (!ormask && !ctx->NeedEyeCoords && !ctx->NeedClipCoords) {
	    if (!ctx->ModelProjectWinMatrixUptodate) 
	       gl_calculate_model_project_win_matrix( ctx );
	    proj_mat = &ctx->ModelProjectWinMatrix;
	    proj_dest = &VB->Win;
	    need_window_transform = 0;
	 }
#endif
      } 
   }


   /* Clip transform.
    */
   VB->ClipPtr = VB->Unprojected;

   if (proj_mat->type != MATRIX_IDENTITY ||
       (ctx->NeedCull && VB->Unprojected->stride != 4*sizeof(GLfloat))) 
   {
      VB->ClipPtr = TransformRaw(proj_dest, proj_mat, VB->Unprojected );
   } 




   /* Cliptest and/or perspective divide.
    */

   if (needclip)
   {
      VB->Projected = gl_clip_tab[VB->ClipPtr->size]( VB->ClipPtr,
						      &VB->Win,
						      VB->ClipMask + VB->Start,
						      &VB->ClipOrMask, 
						      &VB->ClipAndMask );
      if (VB->ClipOrMask) 
	 VB->CullMode |= CLIP_MASK_ACTIVE;      

      if (VB->ClipAndMask) {
	 VB->Culled = 1;	 
	 gl_update_materials(VB);
	 return;      
      }
   } 
   else if (VB->ClipPtr->size == 4) 
   {
      VB->Projected = gl_project_points( &VB->Win, VB->ClipPtr );
   }
   else
   {
      VB->Projected = VB->ClipPtr;
      gl_vector4f_clean_elem(&VB->Win, VB->Count, 3);
   }


   /* User cliptest.
    */
   if (ctx->Transform.AnyClip && !gl_user_cliptest( VB )) {
      VB->Culled = 1;
      gl_update_materials(VB);
      return;
   }
   


   /* Vertex culling - not for cva precalc.
    */
   if (VB->EarlyCull && (VB->ClipOrMask || ctx->NeedCull))
   {
      GLuint cullcount = gl_cull_vb( VB );      
      if (cullcount == VB->Count) { 
	 VB->Culled = 1; 
	 gl_update_materials(VB);
	 return; 
      }
      if (cullcount || ctx->LightTwoSide) VB->CullMode |= CULL_MASK_ACTIVE;
   }
   else if (VB->IM) {
      gl_dont_cull_vb( VB );
   }



   /* Fixup for glDrawArrays.
    */
   if (VB->Type == VB_IMMEDIATE && 
      (VB->ClipOrMask || VB->CopyCount != 0)) 
      import_outstanding_data( VB ); 
   


   /* Window transform.
    */
   if (need_window_transform) {

      if (VB->Start != VB->CopyStart) {
	 ADJ_VEC_STRIDE( VB->Projected, copycount );
	 VB->Win.start = (GLfloat *)VB->Win.data[VB->CopyStart];
      }

      (void) Transform( &VB->Win,
			&ctx->Viewport.WindowMap, 
			VB->Projected, 
			VB->ClipMask + VB->CopyStart,
			VB->CullFlag[1]);

      if (VB->Win.size == 2) 
	 gl_vector4f_clean_elem(&VB->Win, VB->Count, 2);      
   }      
}






static void do_normal_transform( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   GLuint tmp = 0;

   if (VB->Type == VB_CVA_PRECALC) 
      VB->NormalPtr = &ctx->CVA.v.Normal;


   if (VB->CullMode & (COMPACTED_NORMALS|CULL_MASK_ACTIVE)) {
      tmp = 1;
      gl_make_normal_cullmask( VB );
   }
	 
   (ctx->NormalTransform[tmp])(&ctx->ModelView,
			       ctx->vb_rescale_factor,
			       VB->NormalPtr,
			       VB->NormalLengthPtr ? VB->NormalLengthPtr + VB->Start : 0,
			       VB->NormCullStart,
			       VB->store.Normal);
	    
   VB->NormalPtr = VB->store.Normal;
}



static void do_lighting( struct vertex_buffer *VB )
{
   GLubyte flags = VB->CullMode & (CULL_MASK_ACTIVE|COMPACTED_NORMALS);

   if ((flags&CULL_MASK_ACTIVE) && !VB->NormCullStart)
      gl_make_normal_cullmask( VB );

   gl_shade_func_tab[VB->ctx->shade_func_flags | flags]( VB );
}   



static void do_update_materials( struct vertex_buffer *VB )
{
   gl_update_materials( VB );
}   


static void do_texture_0( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   
   if (ctx->Texture.Unit[0].TexGenEnabled) 
      (ctx->Texture.Unit[0].func[VB->CullMode & 0x3])( VB, 0 );

   if (VB->TexCoordPtr[0]->stride != 4*sizeof(GLfloat) ||
       ctx->TextureMatrix[0].type != MATRIX_IDENTITY) 
   {
      VB->TexCoordPtr[0] = Transform( VB->store.TexCoord[0],
				      &ctx->TextureMatrix[0], 
				      VB->TexCoordPtr[0], 
				      VB->ClipMask + VB->Start,
				      VB->CullFlag[0]);
   }
}


static void do_texture_1( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;

   if (ctx->Texture.Unit[1].TexGenEnabled) 
      (ctx->Texture.Unit[1].func[VB->CullMode & 0x3])( VB, 1 );

   if (VB->TexCoordPtr[1]->stride != 4*sizeof(GLfloat) ||
       ctx->TextureMatrix[1].type != MATRIX_IDENTITY) 
   {
      VB->TexCoordPtr[1] = Transform( VB->store.TexCoord[1],
				      &ctx->TextureMatrix[1], 
				      VB->TexCoordPtr[1], 
				      VB->ClipMask + VB->Start,
				      VB->CullFlag[0]);
   }
}




/* Inputs: ArrayElt + (ctx->Flags - ctx->Array.LockPrecalcOutputs)
 * Outputs: nil.
 *
 * Kinda hard to compute the inputs, but once done, 
 */
static void do_merge_and_render( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   ctx->Driver.MergeAndRenderCVA( VB, ctx->CVA.VB );
}


static void do_indirect_render( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   ctx->Driver.RenderVBIndirect( VB, ctx->CVA.VB );
}


static void do_partial_setup( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   ctx->Driver.PartialRasterSetupCVA( VB, 
				      ctx->CVA.pre.new_outputs,
				      ctx->CVA.pre.outputs );
}


/* Done if we need to pull cva data into the immediate struct for some
 * further processing, or if there is no ctx->Driver.MergeAndRenderCVA()
 * implemented.  Inputs: ArrayElt - Outputs: ctx->Array.LockPrecalcOutputs
 *
 * Should only be done if a later stage needs LockPrecalcOutputs.  If 
 */
static void do_cva_merge( struct vertex_buffer *VB )
{
   struct gl_cva *cva = &VB->ctx->CVA;
   gl_merge_cva( VB, cva->VB );
}






/* Inputs: ctx->Flags, Outputs: nil
 * 
 * Cva issues will always have been resolved prior to reaching this point.
 * If we do an early render, we will never call this function.
 */
static void do_render( struct vertex_buffer *VB )
{
   gl_render_vb( VB );   
}


/* With CVA, 'full setup' can be acheived by the partial setup code when
 * only some of the elements have changed.
 */
static void do_full_setup( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;

   if (VB->Type == VB_CVA_PRECALC) {
      ctx->Driver.PartialRasterSetupCVA( VB, 
					 ctx->CVA.pre.new_outputs,
					 ctx->CVA.pre.outputs );
   } else 
      ctx->Driver.RasterSetup( VB, VB->CopyStart, VB->Count );
}



static void check_fog( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   if (ctx->Fog.Enabled && ctx->FogMode==FOG_VERTEX) 
   {
      GLuint flags;

      if (ctx->Visual->RGBAflag) 
	 flags = VERT_EYE|VERT_RGBA;       
      else
	 flags = VERT_EYE|VERT_INDEX;
      
      d->type = PIPE_ANY;
      d->inputs = flags;
      d->outputs = VERT_RGBA;
   }
}


static void check_lighting( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   if (ctx->Light.Enabled) 
   {
      GLuint flags = VERT_NORM|VERT_MATERIAL;
      
      if (ctx->Light.NeedVertices)
	 if (ctx->NeedEyeCoords)
	    flags |= VERT_EYE;
	 else
	    flags |= VERT_OBJ_ANY;

      if (ctx->Light.ColorMaterialEnabled) flags |= VERT_RGBA;
      
      d->type = PIPE_ANY;
      d->inputs = flags;
      d->outputs = VERT_RGBA;
   } 
}

static void check_update_materials( GLcontext *ctx, 
				    struct gl_pipeline_stage *d )
{
   if (!ctx->Light.Enabled) {
      GLuint flags = VERT_MATERIAL;
      if (ctx->Light.ColorMaterialEnabled) flags |= VERT_RGBA;
      d->type = PIPE_CVA_ELT|PIPE_IMMEDIATE;
      d->inputs = flags;
      d->outputs = 0;
   }
}

      
static void check_normal_transform( GLcontext *ctx, 
				    struct gl_pipeline_stage *d )
{
   if (ctx->NormalTransform) {
      d->type = PIPE_ANY;
      d->inputs = VERT_NORM;
      d->outputs = VERT_NORM;
   }
}

static void check_texture( GLcontext *ctx, GLuint i, 
			   struct gl_pipeline_stage *d )
{
   struct gl_texture_unit *texUnit = &ctx->Texture.Unit[i];
   
   if (texUnit->Enabled)
   {
      GLuint texflag = VERT_TEX_ANY(i);
      GLuint flags = 0;

      /* safe?
       */
      if (texUnit->GenFlags & TEXGEN_NEED_VERTICES) {
	 if (0)
	    flags |= VERT_OBJ_ANY;
	 if (1)
	    flags |= VERT_EYE;
      }
      
      if (texUnit->GenFlags & TEXGEN_NEED_NORMALS)
	 flags |= VERT_NORM;

      if (texUnit->Enabled & ~texUnit->TexGenEnabled)
	 flags |= texflag;

      d->type = PIPE_ANY;
      d->inputs = flags;
      d->outputs = texflag;
   }
}


static void check_texture_0( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   check_texture(ctx, 0, d);
}

static void check_texture_1( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   check_texture(ctx, 1, d);
}

static void check_cva_merge( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   (void) ctx;
   d->type = PIPE_CVA_ELT;
   d->inputs = VERT_ELT;
   d->outputs = 0; 
}



static void check_render( GLcontext *ctx, struct gl_pipeline_stage *d )
{  
   d->type = PIPE_CVA_ELT|PIPE_IMMEDIATE;
   if (ctx->RenderMode == GL_RENDER && ctx->Driver.RasterSetup)
      d->inputs = VERT_RAST_SETUP_FULL;
   else 
      d->inputs = ctx->RenderFlags; 
   d->outputs = 0;
}


static void check_full_setup( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   if (ctx->RenderMode == GL_RENDER && ctx->Driver.RasterSetup) {
      d->type = PIPE_ANY;
      d->inputs = ctx->RenderFlags;
      d->outputs = VERT_RAST_SETUP_FULL;

      if ( ctx->IndirectTriangles & DD_TRI_CULL )
	 d->type = PIPE_IMMEDIATE|PIPE_CVA_ELT;
   }
}  

static void check_indirect_render( GLcontext *ctx, 
				   struct gl_pipeline_stage *d )
{   
   /* Could be more clever about culling.
    */
   if (ctx->RenderMode == GL_RENDER && !ctx->IndirectTriangles) 
   {
      d->type = PIPE_CVA_ELT;
      d->inputs = VERT_RAST_SETUP_FULL | VERT_ELT;
   }      
}

static void check_merge_and_render( GLcontext *ctx, 
				    struct gl_pipeline_stage *d )
{   
   if (ctx->Driver.CheckMergeAndRenderCVA && !ctx->IndirectTriangles) 
   {
      d->update_inputs = ctx->Driver.CheckMergeAndRenderCVA;
      d->update_inputs( ctx, d );
   }
}

static void check_partial_setup( GLcontext *ctx, struct gl_pipeline_stage *d )
{   
   if (ctx->Driver.CheckPartialRasterSetupCVA && !ctx->IndirectTriangles) 
   {
      d->update_inputs = ctx->Driver.CheckPartialRasterSetupCVA;
      d->update_inputs( ctx, d );
   }
}



static void check_vertex( GLcontext *ctx, struct gl_pipeline_stage *d )
{
   (void) ctx;
   d->type = PIPE_ANY;
   d->inputs = VERT_OBJ_ANY;
   d->outputs = VERT_EYE | VERT_WIN;
}

static void check_cva_prepare_arrays( GLcontext *ctx, 
				      struct gl_pipeline_stage *d )
{
   (void) ctx;
   d->type = PIPE_CVA_PRECALC;
   d->inputs = 0;		/* always run */
   d->outputs = 0;
}


/* 
 */
#define DYN_STATE 0,0,0

static CONST struct gl_pipeline_stage static_pipeline[] = {
   { "cva merge", 
     0,
     PIPE_CVA_ELT,		
     0,
     NEW_CLIENT_STATE,		/* state change */
     0,				/* cva state change */
     0,				/* cva forbidden */
     DYN_STATE,
     check_cva_merge,
     do_cva_merge },
   
   { "cva prepare arrays",
     0,
     PIPE_CVA_PRECALC,
     0,
     0,
     NEW_CLIENT_STATE,
     0,
     PIPE_CVA_PRECALC, 0, 0,
     check_cva_prepare_arrays,
     gl_prepare_arrays_cva },

   { "vertex pipeline",
     PIPE_OP_VERT_XFORM,
     PIPE_CVA_PRECALC|PIPE_IMMEDIATE,
     0,
     0,
     NEW_MODELVIEW|NEW_PROJECTION|NEW_USER_CLIP,
     0,
     0,VERT_OBJ_ANY, VERT_EYE|VERT_WIN,
     check_vertex,
     do_vertex_pipeline },
   
   { "normal transform",
     PIPE_OP_NORM_XFORM,
     PIPE_ANY,
     0,
     NEW_NORMAL_TRANSFORM,
     NEW_NORMAL_TRANSFORM,
     0,
     DYN_STATE,
     check_normal_transform,
     do_normal_transform },
   
   { "lighting", 
     PIPE_OP_LIGHT,
     PIPE_ANY,
     0,
     NEW_LIGHTING,
     NEW_LIGHTING|NEW_MODELVIEW,
     VERT_MATERIAL,		/* I hate glMaterial() */
     DYN_STATE,
     check_lighting,
     do_lighting },
   
   { "update materials (no lighting)",
     0,
     PIPE_CVA_ELT|PIPE_IMMEDIATE,
     0,
     NEW_LIGHTING,
     0,
     0,
     DYN_STATE,
     check_update_materials,
     do_update_materials },
   
   { "fog",
     PIPE_OP_FOG,
     PIPE_ANY,
     0,
     NEW_LIGHTING|NEW_RASTER_OPS,
     NEW_LIGHTING|NEW_RASTER_OPS|NEW_FOG|NEW_MODELVIEW,
     0,
     DYN_STATE,
     check_fog,
     gl_fog_vertices },
   
   { "texture gen/transform 0",
     PIPE_OP_TEX0,
     PIPE_ANY,
     0,
     NEW_TEXTURING,
     NEW_TEXTURING|NEW_TEXTURE_MATRIX,
     0,
     DYN_STATE,
     check_texture_0,
     do_texture_0 },
   
   { "texture gen/transform 1",
     PIPE_OP_TEX1,
     PIPE_ANY,
     0,
     NEW_TEXTURING,
     NEW_TEXTURING|NEW_TEXTURE_MATRIX,
     0,
     DYN_STATE,
     check_texture_1,
     do_texture_1 },   

   { "cva indirect render",
     PIPE_OP_RENDER,
     PIPE_CVA_ELT,
     0,
     NEW_LIGHTING|NEW_TEXTURING|NEW_RASTER_OPS,     
     0,
     0,
     DYN_STATE,
     check_indirect_render,
     do_indirect_render },

   { "cva merge & render", 
     PIPE_OP_RAST_SETUP_1|PIPE_OP_RENDER, /* rsetup1 if processing in elt? */
     PIPE_CVA_ELT,		
     0,
     NEW_CLIENT_STATE,  
     0,				/* cva state change */     
     0,				/* cva forbidden */
     DYN_STATE,
     check_merge_and_render,
     do_merge_and_render },

   { "full raster setup",	/* prepare for indirect and normal render */
     PIPE_OP_RAST_SETUP_0|PIPE_OP_RAST_SETUP_1,
     PIPE_ANY,
     0,
     NEW_LIGHTING|NEW_TEXTURING|NEW_RASTER_OPS|NEW_POLYGON,     
     NEW_LIGHTING|NEW_TEXTURING|NEW_RASTER_OPS|NEW_POLYGON,
     0,
     DYN_STATE,
     check_full_setup,
     do_full_setup },

   { "partial raster setup",	/* prepare for merge_and_render */
     PIPE_OP_RAST_SETUP_0,
     PIPE_CVA_PRECALC,
     0,
     NEW_LIGHTING|NEW_TEXTURING|NEW_RASTER_OPS|NEW_POLYGON,     
     NEW_LIGHTING|NEW_TEXTURING|NEW_RASTER_OPS|NEW_POLYGON,
     0,
     DYN_STATE,
     check_partial_setup,
     do_partial_setup },

   { "render",
     PIPE_OP_RENDER,
     PIPE_CVA_ELT|PIPE_IMMEDIATE,
     0,
     NEW_LIGHTING|NEW_TEXTURING|NEW_RASTER_OPS,     
     0,
     0,
     DYN_STATE,
     check_render,
     do_render }
};


static const char *pipeline_name[9] = {
   0,
   "Immediate",
   "CVA Elt",
   0,
   "CVA Precalc",
   0,
   0,
   0,
   "CVA Precalc Update"
};

static void pipeline_ctr( struct gl_pipeline *p, GLcontext *ctx, GLuint type )
{
   GLuint i;
   (void) ctx;

   p->state_change = 0;
   p->cva_state_change = 0;
   p->inputs = 0;
   p->outputs = 0;
   p->type = type;
   p->ops = 0;

   for (i = 0 ; i < Elements(static_pipeline) ; i++) 
      p->state_change |= static_pipeline[i].state_change;
}


void gl_pipeline_init( GLcontext *ctx )
{
   MEMCPY( ctx->PipelineStage, static_pipeline, sizeof(static_pipeline) );

   pipeline_ctr( &ctx->Pipeline, ctx, PIPE_IMMEDIATE);
   pipeline_ctr( &ctx->CVA.elt, ctx, PIPE_CVA_ELT );
   pipeline_ctr( &ctx->CVA.pre, ctx, PIPE_CVA_PRECALC );
}


void gl_build_immediate_pipeline( GLcontext *ctx ) 
{
   struct gl_pipeline *p = &ctx->Pipeline;
   struct gl_pipeline_stage *pipeline = ctx->PipelineStage;
   GLuint generated = 0;
   GLuint newstate = ctx->NewState;
   GLuint flags = VERT_WIN;
   GLuint i;
  
   /* Precalculate this value because lots of stages use it.
    */
   if (ctx->RenderMode==GL_FEEDBACK) 
   {
      flags = (VERT_WIN|VERT_RGBA|VERT_INDEX|
	       VERT_NORM|VERT_EDGE|
	       VERT_TEX0_ANY|VERT_TEX1_ANY);
   }

   if (ctx->Texture.Enabled & 0xf)
      flags |= VERT_TEX0_ANY;

   if (ctx->Texture.Enabled & 0xf0)
      flags |= VERT_TEX1_ANY;
   
   if (ctx->Visual->RGBAflag) 
      flags |= VERT_RGBA;
   else 
      flags |= VERT_INDEX;

   if (ctx->Polygon.Unfilled) 
      flags |= VERT_EDGE;
 
   ctx->RenderFlags = flags;
   p->inputs = 0;

   for (i = 0 ; i < Elements(static_pipeline) ; i++) {
      
      if (pipeline[i].state_change & newstate) {
	 pipeline[i].type = 0;
	 pipeline[i].update_inputs(ctx, &pipeline[i] );
      }
      
      pipeline[i].active &= ~PIPE_IMMEDIATE;

      if ((pipeline[i].type & PIPE_IMMEDIATE) && pipeline[i].inputs)
      {
	 pipeline[i].active |= PIPE_IMMEDIATE;
	 p->inputs |= pipeline[i].inputs & ~generated;
	 generated |= pipeline[i].outputs;
	 p->ops |= pipeline[i].ops;

	 /* only one render available to the immediate pipeline - no need
	  * for this check:
	  * if (pipeline[i].ops & PIPE_OP_RENDER)
	  *     break;
	  */
      }
   }
}

#ifndef MESA_VERBOSE
int MESA_VERBOSE = 0
/*  | VERBOSE_PIPELINE */
/*  | VERBOSE_TEXTURE */
/*  | VERBOSE_VARRAY */
/*  | VERBOSE_IMMEDIATE */
/*  | VERBOSE_DRIVER */
;
#endif

/* Called only when we want to use the cva pipelines and there has been
 * a state change.
 */
void gl_update_cva_pipelines( GLcontext *ctx )
{
   struct gl_pipeline_stage *pipeline = ctx->PipelineStage;
   struct gl_cva *cva = &ctx->CVA;
   struct gl_pipeline *pre = &cva->pre;   
   struct gl_pipeline *elt = &cva->elt;
   GLuint i;
   GLuint newstate = ctx->Array.NewState;
   GLuint newinputs = ctx->Array.NewArrayState;
   GLuint changed_ops = 0;
   GLuint active_ops = 0;
   GLuint oldoutputs = pre->outputs;

   pre->outputs = ctx->Array.Summary;
   pre->cva_state_change = 0;
   pre->forbidden_inputs = 0;
   pre->new_inputs = newinputs;


   if (MESA_VERBOSE&VERBOSE_PIPELINE) {
      gl_print_vert_flags("new inputs", newinputs);
      gl_print_vert_flags("cva->orflag", cva->orflag);
      gl_print_vert_flags("array state", ctx->Array.Summary);
   }

   /* Slight hack for lighting.
    */
   if (!(cva->orflag & VERT_MATERIAL))
      pre->outputs |= VERT_MATERIAL;
   else if (oldoutputs & VERT_MATERIAL)
      newinputs |= VERT_MATERIAL;

   cva->orflag = 0;
   if (MESA_VERBOSE&VERBOSE_PIPELINE) gl_print_vert_flags("pre->outputs", pre->outputs);
      
   /* If something changes in the pipeline, tag all subsequent stages
    * using this value for recalcuation.  Also used to build the full
    * pipeline by setting newstate and newinputs to ~0.
    *
    * Because all intermediate values are buffered, the new inputs
    * are enough to fully specify what needs to be calculated, and a
    * single pass identifies all stages requiring recalculation.
    */
   for (i = 0 ; i < Elements(static_pipeline) ; i++) {
      pipeline[i].active &= ~PIPE_CVA_RUN_PRECALC;

      if (pipeline[i].special & PIPE_CVA_PRECALC) {
	 pipeline[i].type = 0;
	 pipeline[i].update_inputs(ctx, &pipeline[i]);
      }

      if (pipeline[i].type & PIPE_CVA_PRECALC) 
      {
	 if ((newstate & pipeline[i].cva_state_change) ||
	     (newinputs & pipeline[i].inputs))
	 {	    
	    changed_ops |= pipeline[i].ops;
	    newinputs |= pipeline[i].outputs;
	    pipeline[i].active &= ~PIPE_CVA_PRECALC;
	    if ((pipeline[i].inputs & pre->outputs) == pipeline[i].inputs &&
		(pipeline[i].ops & active_ops) == 0)
	    {
	       pipeline[i].active |= PIPE_CVA_PRECALC | PIPE_CVA_RUN_PRECALC;
	    } 
	 }
      
	 pre->outputs &= ~pipeline[i].outputs;
	 if (pipeline[i].active & PIPE_CVA_PRECALC) {
	    active_ops |= pipeline[i].ops;
	    pre->outputs |= pipeline[i].outputs;
	    pre->forbidden_inputs |= pipeline[i].cva_forbidden_inputs;

/*  	    printf("after %s, ", pipeline[i].name);  */
/*  	    gl_print_vert_flags("pre->outputs", pre->outputs); */
	 }
      }
      else 
	 pipeline[i].active &= ~PIPE_CVA_PRECALC;
   }

   pre->ops = active_ops;
   pre->new_outputs = newinputs & pre->outputs;

   if (!cva->VB) {
      cva->VB = gl_vb_create_for_cva( ctx, ctx->Const.MaxArrayLockSize );
      gl_alloc_cva_store( cva, ctx->Const.MaxArrayLockSize );
   }

   if (!(pre->outputs & (VERT_RAST_SETUP_FULL|VERT_RAST_SETUP_PART)) && 
       ctx->Driver.InvalidateRasterSetupCVA)
   {
/*        printf("invalidate raster setup\n"); */
      ctx->Driver.InvalidateRasterSetupCVA( cva->VB );
   }

   gl_reset_cva_vb( cva->VB, changed_ops );

   if (1) {
      GLuint available = pre->outputs | VERT_DATA;
      GLuint generated = 0;
      elt->outputs = 0;		/* not used */
      elt->inputs = 0;

      for (i = 0 ; i < Elements(static_pipeline) ; i++) {
	 pipeline[i].active &= ~PIPE_CVA_ELT;

	 /* fix me */
	 if (1) {
	    pipeline[i].type = 0;
	    pipeline[i].update_inputs(ctx, &pipeline[i]);
	 }

	 if ((pipeline[i].type & PIPE_CVA_ELT) &&
	     (pipeline[i].ops & active_ops) == 0 &&
	     (pipeline[i].inputs & available) == pipeline[i].inputs) 
	 {
	    elt->inputs |= pipeline[i].inputs & ~generated;
	    pipeline[i].active |= PIPE_CVA_ELT;
	    generated |= pipeline[i].outputs;
	    available |= pipeline[i].outputs;
	    active_ops |= pipeline[i].ops;
	 
	    if (pipeline[i].ops & PIPE_OP_RENDER) {
	       break;
	    }
	 }
      }

      /* clean the remaining elements */
      for (i++ ; i < Elements(static_pipeline) ; i++) 
	 pipeline[i].active &= ~PIPE_CVA_ELT;

      cva->merge = elt->inputs & pre->outputs;
      elt->ops = active_ops & ~pre->ops;
   }

   if (MESA_VERBOSE&VERBOSE_PIPELINE) {
      gl_print_pipeline( ctx, pre ); 
      gl_print_pipeline( ctx, elt ); 
   }

   cva->pre.data_valid = 0;
   cva->pre.pipeline_valid = 1;
   ctx->Array.NewState = 0;
   ctx->Array.NewArrayState = 0;
}



void gl_update_pipelines( GLcontext *ctx )
{
   GLuint newstate = ctx->NewState;
   struct gl_cva *cva = &ctx->CVA;

   if (ctx->Pipeline.state_change & newstate) {
      gl_build_immediate_pipeline( ctx );
      ctx->Flags = ctx->Pipeline.inputs;
   }
   
   ctx->Array.NewState |= newstate;

   cva->pre.forbidden_inputs = 0;
   cva->pre.pipeline_valid = 0;
   cva->pre.data_valid = 0;
}


void gl_run_pipeline( struct vertex_buffer *VB, GLuint type )
{
   struct gl_pipeline_stage *pipeline = VB->ctx->PipelineStage;
   GLuint i;

   START_FAST_MATH;

   for (i = 0 ; i < Elements(static_pipeline) ; i++) 
      if (pipeline[i].active & type) {
	 pipeline[i].func( VB );	    
	 if (VB->Culled) break;	 
      }
   
   END_FAST_MATH;
}

void gl_print_vert_flags( const char *name, GLuint flags ) 
{
   fprintf(stderr, 
	   "%s: (0x%x) %s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
	   name,
	   flags,
	   (flags & VERT_OBJ_ANY)    ? "vertices, " : "",
	   (flags & VERT_ELT)        ? "array-elt, " : "",
	  (flags & VERT_RGBA)       ? "colors, " : "",
	  (flags & VERT_NORM)       ? "normals, " : "",
	  (flags & VERT_INDEX)      ? "index, " : "",
	  (flags & VERT_EDGE)       ? "edgeflag, " : "",
	  (flags & VERT_MATERIAL)   ? "material, " : "",
	  (flags & VERT_TEX0_ANY)   ? "texcoord0, " : "",
	  (flags & VERT_TEX1_ANY)   ? "texcoord1, " : "",
	  (flags & VERT_EVAL_ANY)   ? "eval-coord, " : "",
	  (flags & VERT_EYE)        ? "eye, " : "",
	  (flags & VERT_WIN)        ? "win, " : "",
	  (flags & VERT_RAST_SETUP_FULL) ? "driver-data, " : "", 
	  (flags & VERT_RAST_SETUP_PART) ? "partial-driver-data, " : ""
         );
}


void gl_print_pipeline( GLcontext *ctx, struct gl_pipeline *p )
{
   struct gl_pipeline_stage *pipeline = ctx->PipelineStage;
   GLuint i;

   fprintf(stderr,"Type: %s\n", pipeline_name[p->type]);

   gl_print_vert_flags("Inputs", p->inputs);
   gl_print_vert_flags("Outputs", p->outputs);

   for (i = 0 ; i < Elements(static_pipeline) ; i++) 
      if (pipeline[i].active & p->type) {
	 fprintf(stderr,"%u: %s\n", i, pipeline[i].name);

	 if (p->type == PIPE_CVA_PRECALC &&
	     (pipeline[i].active & PIPE_CVA_RUN_PRECALC))
	    fprintf(stderr,"\t--> needs precalc\n");

	 gl_print_vert_flags("\tinputs", pipeline[i].inputs);
	 gl_print_vert_flags("\toutputs", pipeline[i].outputs);

	 if (p->type == PIPE_CVA_PRECALC && pipeline[i].cva_forbidden_inputs)
	    gl_print_vert_flags("\tforbidden", 
				pipeline[i].cva_forbidden_inputs);
      }
}


