/* $Id: vbrender.c,v 3.19 1999/05/11 17:43:00 keithw Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * Render points, lines, and polygons.  The only entry point to this
 * file is the gl_render_vb() function.  This function is called after
 * the vertex buffer has filled up or a state change has occurred.
 *
 * This file basically only makes calls to the clipping functions and
 * the point, line and triangle rasterizers via the function pointers.
 *    context->Driver.PointsFunc()
 *    context->Driver.LineFunc()
 *    context->Driver.TriangleFunc()
 */


#ifdef PC_HEADER
#include "all.h"
#else
#include <stdio.h>
#include "clip.h"
#include "context.h"
#include "light.h"
#include "lines.h"
#include "macros.h"
#include "matrix.h"
#include "pb.h"
#include "points.h"
#include "types.h"
#include "vb.h"
#include "vbcull.h"
#include "vbrender.h"
#include "vbindirect.h"
#include "xform.h"
#ifdef XFree86Server
#include "GL/xf86glx.h"
#endif
#endif


/*
 * This file implements rendering of points, lines and polygons defined by
 * vertices in the vertex buffer.
 */

/*
 * Render a line segment from VB[v1] to VB[v2] when either one or both
 * endpoints must be clipped.
 */
void gl_render_clipped_line( GLcontext *ctx, GLuint v1, GLuint v2 )
{
   GLuint provoking_vertex = v2;
   struct vertex_buffer *VB = ctx->VB;
   GLubyte mask = VB->ClipMask[v1] | VB->ClipMask[v2];
/*     GLubyte andmask = VB->ClipMask[v1] & VB->ClipMask[v2]; */

   if ((gl_line_clip_tab[VB->ClipPtr->size])( VB, &v1, &v2, mask ))
      ctx->Driver.LineFunc( ctx, v1, v2, provoking_vertex );
/*     else { */
/*        printf("clip-culled, andmask: %x\n", andmask); */
/*     } */
}



/*
 * Compute Z offsets for a polygon with plane defined by (A,B,C,D)
 * D is not needed.
 */
static void offset_polygon( GLcontext *ctx, GLfloat a, GLfloat b, GLfloat c )
{
   GLfloat ac, bc, m;
   GLfloat offset;

   if (c<0.001F && c>-0.001F) {
      /* to prevent underflow problems */
      offset = 0.0F;
   }
   else {
      ac = a / c;
      bc = b / c;
      if (ac<0.0F)  ac = -ac;
      if (bc<0.0F)  bc = -bc;
      m = MAX2( ac, bc );
      /* m = sqrt( ac*ac + bc*bc ); */

      offset = m * ctx->Polygon.OffsetFactor + ctx->Polygon.OffsetUnits;
   }

   ctx->PointZoffset   = ctx->Polygon.OffsetPoint ? offset : 0.0F;
   ctx->LineZoffset    = ctx->Polygon.OffsetLine  ? offset : 0.0F;
   ctx->PolygonZoffset = ctx->Polygon.OffsetFill  ? offset : 0.0F;
}



/*
 * When glPolygonMode() is used to specify that the front/back rendering
 * mode for polygons is not GL_FILL we end up calling this function.
 */
static void unfilled_polygon( GLcontext *ctx,
                              GLuint n, GLuint vlist[],
                              GLuint pv, GLuint facing )
{
   GLenum mode = facing ? ctx->Polygon.BackMode : ctx->Polygon.FrontMode;
   struct vertex_buffer *VB = ctx->VB;

   if (mode==GL_POINT) {
      GLuint i, j;
      GLboolean edge;

      if (!ctx->Driver.PointsFunc)
         gl_set_point_function( ctx );

      edge = VB->EdgeFlagPrimitive;

      for (i=0;i<n;i++) {
         j = vlist[i];
         if (edge || VB->EdgeFlagPtr->data[j]) {
            (*ctx->Driver.PointsFunc)( ctx, j, j );
         }
      }
   }
   else if (mode==GL_LINE) {
      GLuint i, j0, j1;
      GLboolean edge;

      if (!ctx->Driver.LineFunc)
         gl_set_line_function( ctx );

      ctx->StippleCounter = 0;

      edge = VB->EdgeFlagPrimitive;

      /* draw the edges */
      for (i=0;i<n;i++) {
         j0 = (i==0) ? vlist[n-1] : vlist[i-1];
         j1 = vlist[i];
         if (edge || VB->EdgeFlagPtr->data[j0]) {
            (*ctx->Driver.LineFunc)( ctx, j0, j1, pv );
         }
      }
   }
   else {
      /* Fill the polygon */
      GLuint j0, i;
      j0 = vlist[0];
      for (i=2;i<n;i++) {
         (*ctx->Driver.TriangleFunc)( ctx, j0, vlist[i-1], vlist[i], pv );
      }
   }
}


/*
 * Compute signed area of the n-sided polgyon specified by vertices vb->Win[]
 * and vertex list vlist[].
 * A clockwise polygon will return a negative area.
 * A counter-clockwise polygon will return a positive area.
 */
static GLfloat polygon_area( const struct vertex_buffer *vb,
                             GLuint n, const GLuint vlist[] )
{
   GLfloat area = 0.0F;
   GLuint i;
   for (i=0;i<n;i++) {
      /* area = sum of trapezoids */
      GLuint j0 = vlist[i];
      GLuint j1 = vlist[(i+1)%n];
      GLfloat x0 = vb->Win.data[j0][0];
      GLfloat y0 = vb->Win.data[j0][1];
      GLfloat x1 = vb->Win.data[j1][0];
      GLfloat y1 = vb->Win.data[j1][1];
      GLfloat trapArea = (x0-x1)*(y0+y1);  /* Note: no divide by two here! */
      area += trapArea;
   }
   return area;     /* KW: no divide here either... */
}





/*
 * Render a polygon in which at least one vertex has to be clipped.
 * Input:  n - number of vertices
 *         vlist - list of vertices in the polygon.
 *                 CCW order = front facing.
 *
 * 
 * KW: Currently the only way to do late culling.  This will have
 *     to change for cva.
 */
void gl_render_clipped_triangle( GLcontext *ctx, GLuint n, GLuint vlist[], 
				 GLuint pv )
{
   struct vertex_buffer *VB = ctx->VB;
   GLfloat (*win)[4] = VB->Win.data;
   GLubyte mask = 0;
   GLuint i;

   for (i = 0 ; i < n ; i++) 
      mask |= VB->ClipMask[vlist[i]];

   /* Clip against view volume in clip coord space */
   n = (gl_poly_clip_tab[VB->ClipPtr->size])( VB, n, vlist, mask );
   

   if (n >= 3)
   {
      GLuint facing;   /* 0=front, 1=back */
      GLfloat area = polygon_area( VB, n, vlist );

      if (area==0.0F && !ctx->Polygon.Unfilled) {
         /* polygon has zero area, don't draw it */
         return;
      }

      facing = (area<0.0F) ^ ctx->Polygon.FrontBit;

      if (!((facing+1) & ctx->Polygon.CullBits)) {
         return;   /* culled */
      }

      if (ctx->Polygon.OffsetAny) {
         /* compute plane equation of polygon, apply offset */
         GLuint j0 = vlist[0];
         GLuint j1 = vlist[1];
         GLuint j2 = vlist[2];
         GLuint j3 = vlist[ (n==3) ? 0 : 3 ];
         GLfloat ex = win[j1][0] - win[j3][0];
         GLfloat ey = win[j1][1] - win[j3][1];
         GLfloat ez = win[j1][2] - win[j3][2];
         GLfloat fx = win[j2][0] - win[j0][0];
         GLfloat fy = win[j2][1] - win[j0][1];
         GLfloat fz = win[j2][2] - win[j0][2];
         GLfloat a = ey*fz-ez*fy;
         GLfloat b = ez*fx-ex*fz;
         GLfloat c = ex*fy-ey*fx;
         offset_polygon( ctx, a, b, c );
      }

      if (ctx->LightTwoSide) {
	 VB->Specular = VB->Spec[facing];
	 VB->ColorPtr = VB->Color[facing];
	 VB->IndexPtr = VB->Index[facing];
      }

      /* Render the polygon! */
      if (ctx->Polygon.Unfilled) {
         unfilled_polygon( ctx, n, vlist, pv, facing );
      }
      else {
         /* Draw filled polygon as a triangle fan */
         GLuint i;
         GLuint j0 = vlist[0];
         for (i=2;i<n;i++) {
	    ctx->Driver.TriangleFunc( ctx, j0, vlist[i-1], vlist[i], pv ); 
         }
      }
   }
}

/* Implements 
 */
static void render_triangle( GLcontext *ctx,
                             GLuint v0, GLuint v1, GLuint v2, GLuint pv )
{
   struct vertex_buffer *VB = ctx->VB;
   GLfloat ex, ey, fx, fy, c;
   GLfloat (*win)[4] = VB->Win.data;
   GLuint facing = ((VB->CullMask[v2]&PRIM_FACE_REAR)?1:0);

   if (ctx->Polygon.OffsetAny) {
      /* finish computing plane equation of polygon, compute offset */
      GLfloat fz = win[v2][2] - win[v0][2];
      GLfloat ez = win[v1][2] - win[v0][2];
      GLfloat a;
      GLfloat b;
      ex = win[v1][0] - win[v0][0];
      ey = win[v1][1] - win[v0][1];
      fx = win[v2][0] - win[v0][0];
      fy = win[v2][1] - win[v0][1];
      a = ey*fz-ez*fy;
      b = ez*fx-ex*fz;
      c = ex*fy-ey*fx;
      offset_polygon( ctx, a, b, c );
   }

   if (ctx->LightTwoSide) {
      VB->Specular = VB->Spec[facing];
      VB->ColorPtr = VB->Color[facing];
      VB->IndexPtr = VB->Index[facing];
   }

   if (ctx->Polygon.Unfilled) {
      GLuint vlist[3];
      vlist[0] = v0;
      vlist[1] = v1;
      vlist[2] = v2;
      unfilled_polygon( ctx, 3, vlist, pv, facing );
   }
   else {
      (*ctx->Driver.TriangleFunc)( ctx, v0, v1, v2, pv );
   }
}



/* Implements quad rendering when direct_triangles is false.
 *
 * Need to implement cull check ?
 */
static void render_quad( GLcontext *ctx, GLuint v0, GLuint v1,
                         GLuint v2, GLuint v3, GLuint pv )
{
   struct vertex_buffer *VB = ctx->VB;

   GLuint facing = ((VB->CullMask[pv]&PRIM_FACE_REAR)?1:0);


   if (ctx->Polygon.OffsetAny) {
      GLfloat (*win)[4] = VB->Win.data;
      GLfloat ez = win[v2][2] - win[v0][2];
      GLfloat fz = win[v3][2] - win[v1][2];
      GLfloat ex = win[v2][0] - win[v0][0];
      GLfloat ey = win[v2][1] - win[v0][1];
      GLfloat fx = win[v3][0] - win[v1][0];
      GLfloat fy = win[v3][1] - win[v1][1];
      GLfloat a = ey*fz-ez*fy;
      GLfloat b = ez*fx-ex*fz;
      GLfloat c = ex*fy-ey*fx;
      offset_polygon( ctx, a, b, c );
   }


   if (ctx->LightTwoSide) {
      VB->Specular = VB->Spec[facing];
      VB->ColorPtr = VB->Color[facing];
      VB->IndexPtr = VB->Index[facing];
   }


   /* Render the quad! */
   if (ctx->Polygon.Unfilled) {
      GLuint vlist[4];
      vlist[0] = v0;
      vlist[1] = v1;
      vlist[2] = v2;
      vlist[3] = v3;
      unfilled_polygon( ctx, 4, vlist, pv, facing );
   }
   else {
      (*ctx->Driver.QuadFunc)( ctx, v0, v1, v2, v3, pv );
   }
}




static void render_vb_noop( struct vertex_buffer *VB,
			    GLuint start,
			    GLuint count,
			    GLuint parity )
{
   /* Handles completely culled primitives - not much to do.
    */
   (void) VB;
   (void) start;
   (void) count;
   (void) parity;
}



extern const char *gl_prim_name[];

static GLuint const_vlist[VB_SIZE];


/*
 * Either the vertex buffer is full (VB->Count==VB_MAX) or glEnd() has been
 * called.  Render the primitives defined by the vertices and reset the
 * buffer.
 *
 * This function won't be called if the device driver implements a
 * RenderVB() function.
 */

#define FLUSH_INIT_PB(prim)				\
do {							\
   if (1||ctx->PB->primitive != prim) {			\
      if (ctx->PB->count > 0) gl_flush_pb(ctx);     	\
      PB_INIT( ctx->PB, prim );				\
   }							\
} while(0)


static render_func render_vb_culled[GL_POLYGON+2];
static render_func render_vb_raw[GL_POLYGON+2];


#define TAB    render_vb_culled
#define TAG(x) x##_cull
#define INIT(x)  FLUSH_INIT_PB(x)

#define LOCAL_VARS				\
    GLcontext *ctx = VB->ctx;			\
    const GLubyte *cullmask = VB->CullMask;	\
    GLuint vlist[VB_SIZE];			\
    (void) vlist;

#define RENDER_POINTS( start, count )		\
   (void) cullmask;					\
   (*ctx->Driver.PointsFunc)( ctx, start, count-1 );


#define RENDER_LINE( i1, i )		\
do {						\
   const GLubyte flags = cullmask[i];		\
						\
   if (!(flags & PRIM_NOT_CULLED))		\
      continue;					\
						\
   if (flags & PRIM_ANY_CLIP)			\
      gl_render_clipped_line( ctx, i1, i );	\
   else						\
      ctx->Driver.LineFunc( ctx, i1, i, i );	\
} while (0)



#define RENDER_TRI( i2, i1, i, pv, parity)		\
do {							\
   const GLubyte flags = cullmask[i];			\
							\
   if (!(flags & PRIM_NOT_CULLED))			\
      continue;						\
							\
   if (flags & PRIM_ANY_CLIP) {				\
      if (parity) {					\
	 vlist[0] = i1;					\
	 vlist[1] = i2;					\
	 vlist[2] = i;					\
      } else {						\
	 vlist[0] = i2;					\
	 vlist[1] = i1;					\
	 vlist[2] = i;					\
      }							\
      gl_render_clipped_triangle( ctx, 3, vlist, pv );	\
   } else if (parity) 					\
      ctx->TriangleFunc( ctx, i1, i2, i, pv );		\
   else							\
      ctx->TriangleFunc( ctx, i2, i1, i, pv );		\
							\
} while (0)


#define RENDER_QUAD( i3, i2, i1, i, pv)			\
do {							\
   const GLubyte flags = cullmask[i];			\
							\
   if (!(flags & PRIM_NOT_CULLED))			\
      continue;						\
							\
   if (flags&PRIM_ANY_CLIP) {				\
      vlist[0] = i3;					\
      vlist[1] = i2;					\
      vlist[2] = i1;					\
      vlist[3] = i;					\
      gl_render_clipped_triangle( ctx, 4, vlist, pv );	\
   } else						\
      ctx->QuadFunc( ctx, i3, i2, i1, i, pv );		\
} while (0)




#include "render_tmp.h"








#define TAB    render_vb_raw
#define TAG(x) x##_raw
#define LOCAL_VARS  GLcontext *ctx = VB->ctx
#define INIT(x)  FLUSH_INIT_PB(x);


#define RENDER_POINTS( start, count ) \
   (*ctx->Driver.PointsFunc)( ctx, start, count-1 )


#define RENDER_LINE( i1, i ) \
   (*ctx->Driver.LineFunc)( ctx, i1, i, i )


#define RENDER_TRI( i2, i1, i, pv, parity ) 	\
do {							\
   if (parity)						\
      ctx->TriangleFunc( ctx, i1, i2, i, pv );		\
   else							\
      ctx->TriangleFunc( ctx, i2, i1, i, pv );		\
} while (0)


#define RENDER_QUAD( i3, i2, i1, i, pv )	\
   ctx->QuadFunc( ctx, i3, i2, i1, i, i );




#include "render_tmp.h"










/* Could eventually generalize to handle changes of rasterization
 * state other than change-of-primitive.  An example might be 
 * some bind-texture calls.
 */
void gl_render_vb( struct vertex_buffer *VB )
{
   GLcontext *ctx = VB->ctx;
   static GLboolean edge_flag[GL_POLYGON+2] = { 0,0,0,0,1,0,0,1,0,1,0 };
   GLuint i, next, prim;
   GLuint parity = VB->Parity;
   render_func *tab;
   GLuint count = VB->Count;

   if (VB->Indirect) { 
      ctx->Driver.RenderVBIndirect( VB, VB );  
      return; 
   } else if (VB->CullMode & (CULL_MASK_ACTIVE|CLIP_MASK_ACTIVE)) {
      tab = ctx->Driver.RenderVBCulledTab;
   } else {
      tab = ctx->Driver.RenderVBRawTab;
   }

   if (ctx->Current.Primitive == GL_POLYGON+1 && ctx->Driver.RenderStart)
      ctx->Driver.RenderStart( ctx );

   i = VB->CopyStart;
   
   for ( i= VB->CopyStart ; i < count ; parity = 0, i = next ) 
   {
      prim = VB->Primitive[i];
      next = VB->NextPrimitive[i];
      VB->EdgeFlagPrimitive = edge_flag[prim];

/*       printf("gl_render_vb: prim: %s/%d %d-%d\n", gl_prim_name[prim], parity, i, next ); */

      tab[prim]( VB, i, next, parity );

      if (ctx->LightTwoSide) {
	 VB->Specular = VB->Spec[0];
	 VB->ColorPtr = VB->Color[0];
	 VB->IndexPtr = VB->Index[0];
      }
   }

   if (ctx->PB->count > 0) 
      gl_flush_pb(ctx);

   if (ctx->Current.Primitive == GL_POLYGON+1 && ctx->Driver.RenderFinish)
      ctx->Driver.RenderFinish( ctx );
}


void gl_set_render_vb_function( GLcontext *ctx )
{
   if (ctx->Driver.RenderVBCulledTab == 0)
      ctx->Driver.RenderVBCulledTab = render_vb_culled;

   if (ctx->Driver.RenderVBRawTab == 0)
      ctx->Driver.RenderVBRawTab = render_vb_raw;

   /* Culling will be done earlier by gl_cull_vb().
    */
   if (ctx->IndirectTriangles & ~DD_TRI_CULL) {
      ctx->TriangleFunc = render_triangle;
      ctx->QuadFunc = render_quad;
   } else {
      ctx->TriangleFunc = ctx->Driver.TriangleFunc;
      ctx->QuadFunc = ctx->Driver.QuadFunc;
   }

   if (ctx->Driver.RenderVBIndirect == 0)
      ctx->Driver.RenderVBIndirect = gl_render_vb_indirect;

   if (ctx->Driver.RenderElementsCVA == 0)
      ctx->Driver.RenderElementsCVA = gl_render_elts;
}

void gl_init_vbrender( void )
{
   GLuint i;
   for (i=0;i<VB_SIZE;i++) {
      const_vlist[i] = i;
   }
   render_init_raw();
   render_init_cull();
/*    render_init_indirect(); */
}



